#pragma once
#ifndef __HEADER___COST_APPLI_IMAGE_COMPRESSION__
#define __HEADER___COST_APPLI_IMAGE_COMPRESSION__

#include<CGAL/squared_distance_2.h>
#include<CGAL/Cartesian/line_constructions_2.h>

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include"util_define.h"
#include"util_for_cgal.h"
#include"save_svg.h"

template<typename Ini_data_cost>
struct Cost_vertices_ImageCompression
{
	template<typename Triangulation, typename Vertex, typename DataLabel>
	static void p_label(Triangulation& t, Vertex& v, DataLabel& data_label)
	{
		data_label.label = recompute_label<DataLabel::Label>(t, v);
	}
	template<typename Label, typename Triangulation, typename Vertex>
	static Label recompute_label(Triangulation& t, Vertex& v)
	{
		const auto& p = v.point();
		if (Ini_data_cost::is_not_valid(p))
		{
			int x = 0;
			int y = 0;
			if (p.x() <= 0 && p.y() <= 0)
				x = 0;
			else if (p.x() <= 0 && p.y() >= Ini_data_cost::img_color.rows - 1)
				y = Ini_data_cost::img_color.rows - 1;
			else if (p.x() >= Ini_data_cost::img_color.cols - 1 && p.y() <= 0)
				x = Ini_data_cost::img_color.cols - 1;
			else
			{
				x = Ini_data_cost::img_color.cols - 1;
				y = Ini_data_cost::img_color.rows - 1;
			}
			cv::Vec3f color_tmp_cv = Ini_data_cost::img_color[y][x];
			return glm::vec3(color_tmp_cv[0], color_tmp_cv[1], color_tmp_cv[2]);
		}
		double den = 0;
		glm::dvec3 num = glm::dvec3(0.0, 0.0, 0.0);
		typename Vertex::Face_circulator c = v.incident_faces();
		typename Vertex::Face_circulator c_end = c;
		do
		{
			typename Vertex::Face_circulator::Face f = *c;
			int base_v_idx = 0;
			if (f.vertex(base_v_idx) != v.handle())
			{
				++base_v_idx;
				if (f.vertex(base_v_idx) != v.handle())
					++base_v_idx;
			}
			const typename Vertex::Vertex_handle v1 = f.vertex(base_v_idx);
			const typename Vertex::Vertex_handle v2 = f.vertex((base_v_idx + 1) % 3);
			const typename Vertex::Vertex_handle v3 = f.vertex((base_v_idx + 2) % 3);
			const auto& p1 = v1->point();
			const auto& p2 = v2->point();
			const auto& p3 = v3->point();
			const glm::dvec3& c2 = glm::dvec3(t.vertex_data[v2->info()].label);
			const glm::dvec3& c3 = glm::dvec3(t.vertex_data[v3->info()].label);
			if (p1 == p2 || p1 == p3 || p2 == p3)
				continue;
			const auto dir = p2 - p1;
			const auto dir1 = p3 - p1;
			const auto dir2 = p3 - p2;
			const glm::vec2 dirgl = glm::vec2(dir.x(), dir.y());
			const glm::vec2 dirgl1 = glm::vec2(dir1.x(), dir1.y());
			glm::mat2 inv_coord = glm::inverse(glm::mat2(dirgl, dirgl1));

			const auto dir_norm = dir / sqrt(dir.squared_length());
			const auto& normal = dir_norm.perpendicular(CGAL::CLOCKWISE);
			const auto dir1norm = dir1 / sqrt(dir1.squared_length());
			const auto dir2norm = dir2 / sqrt(dir2.squared_length());
			float cost_calc = 0.0f;
			int nb_pt = 0;
			int beta_end = (int)sqrt(CGAL::squared_distance(CGAL::Line_2<typename Vertex::Face_circulator::Face::Vertex_handle::value_type::Geom_traits>(p1, dir), p3));
			for (int beta = 0; beta < beta_end; ++beta)
			{
				auto tmp_p1 = p1 - beta / (normal * dir1norm) * dir1norm;
				auto tmp_p2 = p2 - beta / (normal * dir2norm) * dir2norm;
				int dst = (int)sqrt((tmp_p2 - tmp_p1).squared_length());
				auto tmp_pt = tmp_p1;
				for (int i = 0; i < dst; ++i)
				{
					const int x = (int)tmp_pt.x();
					const int y = (int)tmp_pt.y();
					if (Ini_data_cost::is_in_img(x, y))
					{
						glm::vec2 deltaX = glm::dvec2((tmp_pt - p1).x(), (tmp_pt - p1).y());
						glm::dvec2 ab = glm::dvec2(inv_coord * deltaX);
						const double ab1 = 1.0 - ab.x - ab.y;
						cv::Vec3f color_tmp_cv = Ini_data_cost::img_color[y][x];
						const glm::dvec3 ref_color = glm::dvec3(color_tmp_cv[0], color_tmp_cv[1], color_tmp_cv[2]);
						den += ab1*ab1;
						num += ab1 * (ref_color - ab.x*c2 - ab.y * c3);
					}
					tmp_pt = tmp_pt + dir_norm;
				}
			}

		} while (++c != c_end);
		if (den != 0)
			return glm::vec3(num * (1.0 / den));
		int x = (int)p.x(), y = (int)p.y();
		if (Ini_data_cost::is_in_img(x, y))
		{
			cv::Vec3f color_tmp_cv = Ini_data_cost::img_color[y][x];
			return glm::vec3(color_tmp_cv[0], color_tmp_cv[1], color_tmp_cv[2]);
		}
		return glm::vec3(0.0f, 0.0f, 0.0f);
	}
	
	template<typename Triangulation, typename Vertex>
	static float cost(Triangulation& t, Vertex& v, decltype(Triangulation::vertex_data)& vertex_data_new = decltype(Triangulation::vertex_data)(), decltype(Triangulation::edge_data)& edge_data_new = decltype(Triangulation::edge_data)(), decltype(Triangulation::face_data)& face_data_new = decltype(Triangulation::face_data)())
	{
		return ((float)Ini_data_cost::alpha)/ ((float)Ini_data_cost::target_nb_point);
	}
};

template<typename Ini_data_cost>
struct Cost_edges_ImageCompression
{
	template<typename Triangulation, typename Edge, typename DataLabel>
	static void p_label(Triangulation& t, Edge& e, DataLabel& data_label)
	{
	}
	template<typename Label, typename Triangulation, typename Edge>
	static Label recompute_label(Triangulation& t, Edge& e)
	{
		return Label();
	}
	template<typename Triangulation, typename Edge>
	static float cost(Triangulation& t, Edge& e, decltype(Triangulation::vertex_data)& vertex_data_new = decltype(Triangulation::vertex_data)(), decltype(Triangulation::edge_data)& edge_data_new = decltype(Triangulation::edge_data)(), decltype(Triangulation::face_data)& face_data_new = decltype(Triangulation::face_data)())
	{
		return 0.0f;
	}
};



template<typename Ini_data_cost>
struct Cost_faces_ImageCompression
{
	template<typename Triangulation, typename Face, typename DataLabel>
	static void p_label(Triangulation& t, Face& f, DataLabel& label)
	{
	}
	
	template<typename Label, typename Triangulation, typename Face>
	static Label recompute_label(Triangulation& t, Face& f)
	{
		return Label();
	}
	template<typename Triangulation, typename Face>
	static float cost(Triangulation& t, Face& f, decltype(Triangulation::vertex_data)& vertex_data_new = decltype(Triangulation::vertex_data)(), decltype(Triangulation::edge_data)& edge_data_new = decltype(Triangulation::edge_data)(), decltype(Triangulation::face_data)& face_data_new = decltype(Triangulation::face_data)())
	{
		const typename Face::Vertex_handle v1 = f.vertex(0);
		const typename Face::Vertex_handle v2 = f.vertex(1);
		const typename Face::Vertex_handle v3 = f.vertex(2);
		const auto& p1 = v1->point();
		const auto& p2 = v2->point();
		const auto& p3 = v3->point();

		//get data attached to the vertices
		decltype(Triangulation::vertex_data)::const_iterator it;
		it = vertex_data_new.find(v1->info());
		auto data_v1 = (it != vertex_data_new.end()) ? it->second : t.vertex_data[v1->info()];
		it = vertex_data_new.find(v2->info());
		auto data_v2 = (it != vertex_data_new.end()) ? it->second : t.vertex_data[v2->info()];
		it = vertex_data_new.find(v3->info());
		auto data_v3 = (it != vertex_data_new.end()) ? it->second : t.vertex_data[v3->info()];
		const auto& c1 = data_v1.label;
		const auto& c2 = data_v2.label;
		const auto& c3 = data_v3.label;

		const auto dir = p2 - p1;
		const auto dir1 = p3 - p1;
		const auto dir2 = p3 - p2;
		const glm::vec2 dirgl = glm::vec2(dir.x(), dir.y());
		const glm::vec2 dirgl1 = glm::vec2(dir1.x(), dir1.y());
		glm::mat2 inv_coord = glm::inverse(glm::mat2(dirgl, dirgl1));
		const auto dir_norm = dir / sqrt(dir.squared_length());
		const auto& normal = dir_norm.perpendicular(CGAL::CLOCKWISE);
		const auto dir1norm = dir1 / sqrt(dir1.squared_length());
		const auto dir2norm = dir2 / sqrt(dir2.squared_length());
		float u_fidelity = 0.0f;

		int nb_pt = 0;
#ifdef ROBUST_ERROR_BUT_LONG_TIME
		Triangulation::Locate_type lt;
		int li;
		int row_min = std::max(std::min({ p1.y(), p2.y(), p3.y() }) - 1, 0.0);
		int row_max = std::min(std::max({ p1.y(), p2.y(), p3.y() }) + 1, (double)Ini_data_cost::img_color.rows);
		int col_min = std::max(std::min({ p1.x(), p2.x(), p3.x() }) - 1, 0.0);
		int col_max = std::min(std::max({ p1.x(), p2.x(), p3.x() }) + 1, (double)Ini_data_cost::img_color.cols);
		for (int y = row_min; y < row_max; ++y)
		{
			for (int x = col_min; x < col_max; ++x)
			{
				Triangulation::Face_handle fh = t.locate(Triangulation::Point(x + 0.5, y + 0.5), lt, li);
				if (&*fh != &f)
					continue;
				
				//glm::vec2 deltaX = glm::vec2((tmp_pt - p1).x(), (tmp_pt - p1).y()); 
				glm::vec2 deltaX = glm::vec2((double)x - p1.x() + 0.5, (double)y - p1.y() + 0.5);
				glm::vec2 ab = inv_coord * deltaX;
				const glm::vec3 col = c1 + ab.x * (c2 - c1) + ab.y * (c3 - c1);
				const cv::Vec3f& col_cv_tmp = Ini_data_cost::img_color[y][x];
				const glm::vec3 colcv = glm::vec3(col_cv_tmp[0], col_cv_tmp[1], col_cv_tmp[2]);
				u_fidelity += glm::distance2(col, colcv);
				++nb_pt;
			}

		}
#else
		int beta_end = (int)sqrt(CGAL::squared_distance(CGAL::Line_2<typename Face::Vertex_handle::value_type::Geom_traits>(p1, dir), p3));

		for (int beta = 0; beta < beta_end; ++beta)
		{
			auto tmp_p1 = p1 - beta / (normal * dir1norm) * dir1norm;
			auto tmp_p2 = p2 - beta / (normal * dir2norm) * dir2norm;
			int dst = (int)sqrt((tmp_p2 - tmp_p1).squared_length());
			auto tmp_pt = tmp_p1;
			for (int i = 0; i < dst; ++i)
			{
				const int x = (int)tmp_pt.x();
				const int y = (int)tmp_pt.y();
				if (Ini_data_cost::is_in_img(x, y))
				{
					glm::vec2 deltaX = glm::vec2((tmp_pt - p1).x(), (tmp_pt - p1).y());
					glm::vec2 ab = inv_coord * deltaX;
					const glm::vec3 col = c1 + ab.x * (c2 - c1) + ab.y * (c3 - c1);
					const cv::Vec3f& col_cv_tmp = Ini_data_cost::img_color[y][x];
					const glm::vec3 colcv = glm::vec3(col_cv_tmp[0], col_cv_tmp[1], col_cv_tmp[2]);
					u_fidelity += glm::distance2(col, colcv);
					++nb_pt;
				}
				tmp_pt = tmp_pt + dir_norm;
			}
		}
#endif
		const double tmp_weight_energy =  1.0 /  (Ini_data_cost::img_color.rows * Ini_data_cost::img_color.cols);
		return (float)((1.0 - Ini_data_cost::alpha) * u_fidelity * tmp_weight_energy);

	}
};

struct ini_data_cost_image_compression
{
	//norm of the gradient of the image: used for sampling point
	static cv::Mat_<float> img_grad;
	//input image: used to obtain rows, cols and other stuff
	static cv::Mat_<cv::Vec3f> img_color;
	//main parameter to balance the fidelity term and the prior term defined in equation 8
	static double alpha;
	//parameter used to save files
	static std::string outut_base_name;
	//parameter used by the simulated annealing to compute the kernel ratio defined in equation 5
	static int target_nb_point;


	//this function is required, it is used by the system when the mark alteration kernel operator is called.
	template<typename SA>
	static void operator_label(SA* sa)
	{
	}
	//this function is required.
	static void process_user_classes_input(std::set<glm::ivec2, less_ivec2>& pixel_label4_1, std::set<glm::ivec2, less_ivec2>& pixel_label4_2)
	{
	}
	//this function is required, it is used by the system to check if the selected point is one of the four boundary point of the triangulation wich are at the exterior of the domain (pt is not necessarly composed by integers.
	template<typename Pt>
	static bool is_not_valid(const Pt& pt)
	{
		return pt.x() < 0 || pt.x() >= img_color.cols;
	}
	//this function is required, it is used by the system to check if a parturbation of a vertex do not put it at the exterior of the image domain
	static bool is_in_img(const int& x, const int& y)
	{
		return x >= 0 && y >= 0 && x < img_color.cols&&y < img_color.rows;
	}

	//this function is required, it is used by the system to initialize the DPP
	//the variables img_grad, img_color and outut_base_name should be initilized
	static void ini_from_file(const std::string base_name, const std::string& full_name)
	{
		cv::Mat ini_input = cv::imread(full_name);
		cv::Mat input;
		cv::Mat out_edge, img_binary_tmp;
		cv::Mat input_grad_x;
		cv::Mat input_grad_y;
		cv::Mat_<float> grad_x;
		cv::Mat_<float> grad_y;

		outut_base_name = base_name;

		ini_input.convertTo(img_color, CV_32FC3);
		img_color *= (1.0 / 255);
		cv::cvtColor(ini_input, input, CV_RGB2GRAY);

		cv::Sobel(input, input_grad_x, CV_16S, 1, 0, 3);
		cv::Sobel(input, input_grad_y, CV_16S, 0, 1, 3);
		img_grad = cv::Mat_<float>(input_grad_y.rows, input_grad_y.cols);
		input_grad_x.convertTo(grad_x, CV_32FC1);
		input_grad_y.convertTo(grad_y, CV_32FC1);
		
		float* ptr_grad_x = grad_x[0];
		float* ptr_grad_y = grad_y[0];
		float* ptr_grad_norm = img_grad[0];
		float max = std::numeric_limits<float>::min();
		for (int col = 0; col < grad_x.cols; ++col)
		{
			for (int row = 0; row < grad_x.rows; ++row)
			{
				const float norm = *ptr_grad_norm = sqrt(*ptr_grad_x**ptr_grad_x + *ptr_grad_y**ptr_grad_y);
				if (norm > max)
					max = norm;
				++ptr_grad_x;
				++ptr_grad_y;
				++ptr_grad_norm;
			}
		}
		if (max != 0)
		{
			img_grad *= 1.0 / max;
		}
		cv::GaussianBlur(img_grad, img_grad, { 7, 7 }, 4, 4);
	}

	//this function is required, it is used by the system to save some results
	template<typename SA>
	static void save_output(const SA& sa, const int rows, const int cols)
	{
		cv::imwrite((outut_base_name + "_input.png").c_str(), img_color * 255.0f);
		save_graph_no_label_vertex_color(outut_base_name, "_vert_output.svg", sa.triangulation, rows, cols);
		save_graph_ply((outut_base_name + "_vert_output.ply").c_str(), sa.triangulation);
		save_graph_line_color_from_vertex((outut_base_name + "_vert_output.png").c_str(), sa.triangulation, rows, cols);
		save_graph_line_color_from_vertex((outut_base_name + "_vert_output.bmp").c_str(), sa.triangulation, rows, cols);
		save_graph_only_vertex_color((outut_base_name + "_vert_output.vertex").c_str(), sa.triangulation, rows, false);
		sa.save_energy((outut_base_name + "_energy.csv").c_str());
	}

	//this function is required, it is called by the system at the begining. You can load specific parameters for your application
	static void load_app_specific_parameters()
	{
		std::string str, str2;
		std::istringstream iss;
		std::ifstream ifs("config/app_image_compression_parameters.conf");
		while (std::getline(ifs, str2).good())
		{
			iss = std::istringstream();//otherwise iss  is not updated (or in a bad state after finisshing to read a line)
			iss.str(str2.c_str());
			std::getline(iss, str, ' ');
			if (str == "alpha")
				iss >> alpha;
		}
		ifs.close();
	}
};




#endif  __HEADER___COST_APPLI_IMAGE_COMPRESSION__

