#pragma once
#ifndef __HEADER___COST_APPLI_OBJECT_CONTOURING__
#define __HEADER___COST_APPLI_OBJECT_CONTOURING__

#include<CGAL/squared_distance_2.h>
#include<CGAL/squared_distance_3.h>
#include<CGAL/Cartesian/line_constructions_2.h>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/ml/ml.hpp>
#include <opencv2/core.hpp>
#include <opencv2/imgproc.hpp>
#include "opencv2/imgcodecs.hpp"
#include <opencv2/highgui.hpp>
#include <opencv2/ml.hpp>
#include"util_define.h"
#include"util_for_cgal.h"
#include"gl_widget.h"
#include"save_svg.h"

struct labelObjectContouring
{
	glm::vec3 color;
	int label;
};
template<typename Ini_data_cost>
struct Cost_vertices_object_contouring
{
	template<typename Triangulation, typename Vertex, typename DataLabel>
	static void p_label(Triangulation& t, Vertex& v, DataLabel& data_label)
	{
	}
	template<typename Label, typename Triangulation, typename Vertex>
	static Label recompute_label(Triangulation& t, Vertex& v)
	{
		return Label();
	}
	template<typename Triangulation, typename Vertex>
	static float cost(Triangulation& t, Vertex& v, decltype(Triangulation::vertex_data)& vertex_data_new = decltype(Triangulation::vertex_data)(), decltype(Triangulation::edge_data)& edge_data_new = decltype(Triangulation::edge_data)(), decltype(Triangulation::face_data)& face_data_new = decltype(Triangulation::face_data)())
	{
		return 0.0f;
	}
};

template<typename Ini_data_cost>
struct Cost_edges_object_contouring
{
	template<typename Edge>
	static double weight_edge_length(Edge& e)
	{
		const double le = sqrt((e.first->vertex((e.second + 2) % 3)->point() - e.first->vertex((e.second + 1) % 3)->point()).squared_length());
		if (le == 0)
			return 10.0;
		constexpr double param_exp = 4.59511985;//=std::log(99.0);
		constexpr double delta = 0.5;
		const double delta_lm = Ini_data_cost::min_length_edge*delta;
		const double delta_lm_2 = Ini_data_cost::min_length_edge*(1.0 - delta);
		return  1.0 / (1.0 + std::exp(-(delta_lm_2 - le) / delta_lm * param_exp));
	}
	template<typename Triangulation, typename Edge, typename DataLabel>
	static void p_label(Triangulation& t, Edge& e, DataLabel& data_label)
	{
	}
	template<typename Label, typename Triangulation, typename Edge>
	static Label recompute_label(Triangulation& t, Edge& e)
	{
		return Label();
	}
	template<typename Triangulation, typename Edge>
	static float cost(Triangulation& t, Edge& e, decltype(Triangulation::vertex_data)& vertex_data_new = decltype(Triangulation::vertex_data)(), decltype(Triangulation::edge_data)& edge_data_new = decltype(Triangulation::edge_data)(), decltype(Triangulation::face_data)& face_data_new = decltype(Triangulation::face_data)())
	{
		//compute w_l of equation 12
		const auto& v1 = e.first->vertex((e.second + 1) % 3);
		const auto& v2 = e.first->vertex((e.second + 2) % 3);
		const auto& p1 = v1->point();
		const auto& p2 = v2->point();
		if (Ini_data_cost::is_not_valid(p1) && Ini_data_cost::is_not_valid(p2))
			return 0.0f;
		if (Ini_data_cost::is_not_valid(p1) || Ini_data_cost::is_not_valid(p2))
			return 0.0f;

		const double w_l = weight_edge_length(e);
		if (w_l == 10)//ie  edge of length null
			return 10.0f;
		return (float)(Ini_data_cost::alpha * Ini_data_cost::mu * w_l);
	}
};
template<typename Ini_data_cost>
struct Cost_faces_object_contouring
{
	template<typename Triangulation, typename Face>
	static std::pair<float, float> p_label(Triangulation& t, Face& f)
	{
		//the p_label used in the function cost is directly computed in the function cost.
		//here the p label is not the same
		const glm::vec3& cfcst = t.face_data[f.info()].label.color;
		int classe = 0;
		double dst_1 = 1000, dst_2 = 1000, tmp_dst;
		for (const glm::vec3& v : Ini_data_cost::color_label4_1)
		{
			if ((tmp_dst = glm::distance2(v, cfcst)) < dst_1)
				dst_1 = tmp_dst;
		}
		for (const glm::vec3& v : Ini_data_cost::color_label4_2)
		{
			if ((tmp_dst = glm::distance2(v, cfcst)) < dst_2)
			{
				dst_2 = tmp_dst;
			}
		}
		dst_1 = sqrt(dst_1);
		dst_2 = sqrt(dst_2);
		const double sum = dst_1 + dst_2;
		if (sum == 0)
			return std::make_pair(0.5f, 0.5f);
		return{ (float)(dst_1 / (dst_1 + dst_2)) ,(float)(dst_2 / (dst_1 + dst_2)) };
	}
	template<typename Triangulation, typename Face, typename DataLabel >
	static void p_label(Triangulation& t, Face& f, DataLabel& data_label)
	{
		data_label.label.color = recompute_new_color_on_face_cst(t, f);
		if (data_label.label.label == 0)
			data_label.p_label = p_label(t, f).first;
		else
			data_label.p_label = p_label(t, f).second;
		//if (data_label.label.label == 0)
		//	return p_label(t, f).first;
		//return p_label(t, f).second;
	}
	template<typename Label, typename Triangulation, typename Face>
	static Label recompute_label(Triangulation& t, Face& f)
	{
		//t.face_data[f.info()].color = recompute_new_color_on_face_cst(t, f);
		const std::pair<float, float> p = p_label(t, f);
		labelObjectContouring tmp;
		tmp.color = recompute_new_color_on_face_cst(t, f);
		if (p.first < p.second)
		{
			tmp.label = 0;
			return tmp;
		}
		tmp.label = 1;
		return  tmp;
	}
	template<typename Triangulation, typename Face>
	static float cost(Triangulation& t, Face& f, decltype(Triangulation::vertex_data)& vertex_data_new = decltype(Triangulation::vertex_data)(), decltype(Triangulation::edge_data)& edge_data_new = decltype(Triangulation::edge_data)(), decltype(Triangulation::face_data)& face_data_new = decltype(Triangulation::face_data)())
	{
		const typename Face::Vertex_handle v1 = f.vertex(0);
		const typename Face::Vertex_handle v2 = f.vertex(1);
		const typename Face::Vertex_handle v3 = f.vertex(2);
		const auto& p1 = v1->point();
		const auto& p2 = v2->point();
		const auto& p3 = v3->point();
		
		//get data attached to the current triangle
		decltype(Triangulation::face_data)::const_iterator itf = face_data_new.find(f.info());
		auto data_f = (itf != face_data_new.end()) ? itf->second : t.face_data[f.info()];
		const int label = data_f.label.label;
		const glm::vec3& cfcst = data_f.label.color;

		//get data attached to the three neighborhood faces
		itf = face_data_new.find(f.neighbor(0)->info());
		auto data_f1 = (itf != face_data_new.end()) ? itf->second : t.face_data[f.neighbor(0)->info()];
		itf = face_data_new.find(f.neighbor(1)->info());
		auto data_f2 = (itf != face_data_new.end()) ? itf->second : t.face_data[f.neighbor(1)->info()];
		itf = face_data_new.find(f.neighbor(2)->info());
		auto data_f3 = (itf != face_data_new.end()) ? itf->second : t.face_data[f.neighbor(2)->info()];

		const auto dir = p2 - p1;
		const auto dir1 = p3 - p1;
		const auto dir2 = p3 - p2;
		const glm::vec2 dirgl = glm::vec2(dir.x(), dir.y());
		const glm::vec2 dirgl1 = glm::vec2(dir1.x(), dir1.y());
		//	glm::mat2 inv_coord = glm::inverse(glm::outerProduct(dirgl, dirgl1));
		glm::mat2 inv_coord = glm::inverse(glm::mat2(dirgl, dirgl1));
		const auto dir_norm = dir / sqrt(dir.squared_length());
		const auto& normal = dir_norm.perpendicular(CGAL::CLOCKWISE);
		const auto dir1norm = dir1 / sqrt(dir1.squared_length());
		const auto dir2norm = dir2 / sqrt(dir2.squared_length());
		float u_fidelity = 0.0f;
		double u_label_2 = 0.0;
		double expected_label_facet = 0.0;


		int nb_pt = 0;
#ifdef ROBUST_ERROR_BUT_LONG_TIME
		Triangulation::Locate_type lt;
		int li;
		int row_min = std::max(std::min({ p1.y(), p2.y(), p3.y() }) - 1, 0.0);
		int row_max = std::min(std::max({ p1.y(), p2.y(), p3.y() }) + 1, (double)ini_data_cost_img::img.rows);
		int col_min = std::max(std::min({ p1.x(), p2.x(), p3.x() }) - 1, 0.0);
		int col_max = std::min(std::max({ p1.x(), p2.x(), p3.x() }) + 1, (double)ini_data_cost_img::img.cols);
		for (int y = row_min; y < row_max; ++y)
		{
			for (int x = col_min; x < col_max; ++x)
			{
				Triangulation::Face_handle fh = t.locate(Triangulation::Point(x + 0.5, y + 0.5), lt, li);
				if (&*fh != &f)
					continue;
				const cv::Vec3f& col_cv_tmp = Ini_data_cost::img_color[y][x];
				const glm::vec3 colcv = glm::vec3(col_cv_tmp[0], col_cv_tmp[1], col_cv_tmp[2]);
				u_fidelity += glm::distance2(cfcst, colcv);
				++nb_pt;
				double tmp_u_label_2 = (label == 0) ? Ini_data_cost::img_proba_label_1[y][x] : 1.0 - Ini_data_cost::img_proba_label_1[y][x];
				u_label_2 += tmp_u_label_2;
			}
		}
#else
		int beta_end = (int)sqrt(CGAL::squared_distance(CGAL::Line_2<typename Face::Vertex_handle::value_type::Geom_traits>(p1, dir), p3));
		for (int beta = 0; beta < beta_end; ++beta)
		{
			auto tmp_p1 = p1 - beta / (normal * dir1norm) * dir1norm;
			auto tmp_p2 = p2 - beta / (normal * dir2norm) * dir2norm;
			int dst = (int)sqrt((tmp_p2 - tmp_p1).squared_length());
			auto tmp_pt = tmp_p1;
			for (int i = 0; i < dst; ++i)
			{
				const int x = (int)tmp_pt.x();
				const int y = (int)tmp_pt.y();
				if (Ini_data_cost::is_in_img(x, y))
				{
					const cv::Vec3f& col_cv_tmp = Ini_data_cost::img_color[y][x];
					const glm::vec3 colcv = glm::vec3(col_cv_tmp[0], col_cv_tmp[1], col_cv_tmp[2]);
					u_fidelity += glm::distance2(cfcst, colcv);
					++nb_pt;
					double tmp_u_label_2 = (label == 0) ? Ini_data_cost::img_proba_label_1[y][x] : 1.0 - Ini_data_cost::img_proba_label_1[y][x];
					u_label_2 += tmp_u_label_2;
				}
				tmp_pt = tmp_pt + dir_norm;
			}
		}
#endif
		//computation of w_s of equation 12: it is the length of the contour of the object
		double w_s = 0.0;
		if (data_f1.label.label != label && !t.is_infinite(f.neighbor(0)))
		{
			w_s += sqrt((p2 - p3).squared_length());
		}
		if (data_f2.label.label != label && !t.is_infinite(f.neighbor(1)))
		{
			w_s += sqrt((p1 - p3).squared_length());
		}
		if (data_f3.label.label != label && !t.is_infinite(f.neighbor(2)))
		{
			w_s += sqrt((p1 - p2).squared_length());
		}
		w_s *= 0.5;//to not count twice edges ...

		const double tmp_weight_energy = Ini_data_cost::target_nb_point * 1.0 / ( Ini_data_cost::img_color.rows * Ini_data_cost::img_color.cols);
		
		w_s *= tmp_weight_energy;//w_s * beta 2

		u_fidelity *= (float)tmp_weight_energy;
		u_label_2 *= tmp_weight_energy;

		return (float)((1.0 - Ini_data_cost::alpha) * (u_label_2 + u_fidelity) + Ini_data_cost::alpha * (1 - Ini_data_cost::mu) * w_s);

	}
	template<typename Triangulation, typename Face>
	static glm::vec3 recompute_new_color_on_face_cst(Triangulation& t, Face& f)
	{
		glm::dvec3 num(0.0f, 0.0f, 0.0f);
		int den = 0;
		const typename Face::Vertex_handle v1 = f.vertex(0);
		const typename Face::Vertex_handle v2 = f.vertex(1);
		const typename Face::Vertex_handle v3 = f.vertex(2);
		const auto& p1 = v1->point();
		const auto& p2 = v2->point();
		const auto& p3 = v3->point();
		const auto dir = p2 - p1;
		const auto dir1 = p3 - p1;
		const auto dir2 = p3 - p2;
		const glm::vec2 dirgl = glm::vec2(dir.x(), dir.y());
		const glm::vec2 dirgl1 = glm::vec2(dir1.x(), dir1.y());
		glm::mat2 inv_coord = glm::inverse(glm::mat2(dirgl, dirgl1));

		const auto dir_norm = dir / sqrt(dir.squared_length());
		const auto& normal = dir_norm.perpendicular(CGAL::CLOCKWISE);
		const auto dir1norm = dir1 / sqrt(dir1.squared_length());
		const auto dir2norm = dir2 / sqrt(dir2.squared_length());
		float cost_calc = 0.0f;
		int nb_pt = 0;
		int beta_end = (int)sqrt(CGAL::squared_distance(CGAL::Line_2<typename Face::Vertex_handle::value_type::Geom_traits>(p1, dir), p3));
		for (int beta = 0; beta < beta_end; ++beta)
		{
			auto tmp_p1 = p1 - beta / (normal * dir1norm) * dir1norm;
			auto tmp_p2 = p2 - beta / (normal * dir2norm) * dir2norm;
			int dst = (int)sqrt((tmp_p2 - tmp_p1).squared_length());
			auto tmp_pt = tmp_p1;
			for (int i = 0; i < dst; ++i)
			{
				const int x = (int)tmp_pt.x();
				const int y = (int)tmp_pt.y();
				if (Ini_data_cost::is_in_img(x, y))
				{
					cv::Vec3f color_tmp_cv = Ini_data_cost::img_color[y][x];
					const glm::dvec3 ref_color = glm::dvec3(color_tmp_cv[0], color_tmp_cv[1], color_tmp_cv[2]);
					num += ref_color;
					++den;
				}
				tmp_pt = tmp_pt + dir_norm;
			}
		}
		if (den != 0)
			return glm::vec3(num * (1.0 / den));
		int x = (int)p1.x();
		int y = (int)p1.y();
		if (Ini_data_cost::is_in_img(x, y))
		{
			cv::Vec3f color_tmp_cv = Ini_data_cost::img_color[y][x];
			return glm::vec3(color_tmp_cv[0], color_tmp_cv[1], color_tmp_cv[2]);
		}
		x = (int)p2.x();
		y = (int)p2.y();
		if (Ini_data_cost::is_in_img(x, y))
		{
			cv::Vec3f color_tmp_cv = Ini_data_cost::img_color[y][x];
			return glm::vec3(color_tmp_cv[0], color_tmp_cv[1], color_tmp_cv[2]);
		}
		x = (int)p3.x();
		y = (int)p3.y();
		if (Ini_data_cost::is_in_img(x, y))
		{
			cv::Vec3f color_tmp_cv = Ini_data_cost::img_color[y][x];
			return glm::vec3(color_tmp_cv[0], color_tmp_cv[1], color_tmp_cv[2]);
		}
		return glm::vec3(0.0f, 0.0f, 0.0f);
	}
};

struct ini_data_cost_object_contouring
{
	//norm of the gradient of the image: used for sampling point
	static cv::Mat_<float> img_grad;
	//input image: used to obtain rows, cols and other stuff
	static cv::Mat_<cv::Vec3f> img_color;
	//main parameter to balance the fidelity term and the prior term defined in equation 8
	static double alpha;
	//parameter used to save files
	static std::string outut_base_name;
	//parameter used by the simulated annealing to compute the kernel ratio defined in equation 5
	static int target_nb_point;
	
	//specific-application variables
	//img_proba_label_1 is the probability map H of equation 11
	//color_label4_j contain the colors of the input image convolved by a mean filter on the position of the pixel labelled as class j by the user. This variable is used to compute H using equation 13
	//min length edge is the parameter b of the sigmoid S_{a,b} of the function w_l of the equation 12
	static cv::Mat_<float> img_proba_label_1;
	static std::vector<glm::vec3> color_label4_1;
	static std::vector<glm::vec3> color_label4_2;
	static int min_length_edge;
	static double mu;

	//this function is required, it is used by the system when the mark alteration kernel operator is called.
	template<typename SA>
	static void operator_label(SA* sa)
	{
		sa->operator_label_partition_face();
	}	
	//this function is required.
	static void process_user_classes_input(std::set<glm::ivec2, less_ivec2 >& pixel_label4_1, std::set<glm::ivec2, less_ivec2>& pixel_label4_2)
	{
		color_label4_1.clear();
		color_label4_2.clear();
		for (const glm::ivec2& px : pixel_label4_1)
			color_label4_1.emplace_back(average(px.x, px.y, 4));
		for (const glm::ivec2& px : pixel_label4_2)
			color_label4_2.emplace_back(average(px.x, px.y, 4));

		float* proba = img_proba_label_1[0];
		cv::Vec3f* color = img_color[0];
		for (int row = 0; row < img_color.rows; ++row)
		{
			for (int col = 0; col < img_color.cols; ++col, ++color, ++proba)
			{
				glm::vec3 cfcst = glm::vec3((*color)[0], (*color)[1], (*color)[2]);
				int classe = 0;
				double dst_1 = 1000, dst_2 = 1000, tmp_dst;
				for (const glm::vec3& v : color_label4_1)
				{
					if ((tmp_dst = glm::distance2(v, cfcst)) < dst_1)
						dst_1 = tmp_dst;
				}
				for (const glm::vec3& v : color_label4_2)
				{
					if ((tmp_dst = glm::distance2(v, cfcst)) < dst_2)
					{
						dst_2 = tmp_dst;
					}
				}
				dst_1 = sqrt(dst_1);
				dst_2 = sqrt(dst_2);
				const double sum = dst_1 + dst_2;
				if (sum == 0)
					*proba = 0.5f;
				else
					*proba = (float)(dst_1 / (dst_1 + dst_2));
			}
		}
		cv::imwrite((outut_base_name + "_proba_map.png").c_str(), img_proba_label_1 * 255.0f);
		cv::imwrite("$$$$bin.png", 255.0f * img_proba_label_1);
	}
	//this function is required, it is used by the system to check if the selected point is one of the four boundary point of the triangulation wich are at the exterior of the domain (pt is not necessarly composed by integers.
	template<typename Pt>
	static bool is_not_valid(const Pt& pt)
	{
		return pt.x() < 0 || pt.x() >= img_color.cols;
	}
	//this function is required, it is used by the system to check if a parturbation of a vertex do not put it at the exterior of the image domain
	static bool is_in_img(const int& x, const int& y)
	{
		return x >= 0 && y >= 0 && x < img_color.cols&&y < img_color.rows;
	}
	//this is a specific appication function. This function compute the average color of the image on the area of size 2*m+1 centerd on (col, row)
	static glm::vec3 average(int col, int row, const int m)
		{
			if (col < 0 || row < 0 || col >= img_color.cols || row >= img_color.rows || m < 0)
				return glm::vec3(0.0f, 0.0f, 0.0f);
			int x = col - m, y = row - m, w = col + m, h = row + m;
			if (x < 0)
				x = 0;
			if (y < 0)
				y = 0;
			if (w >= img_color.cols)
				w = img_color.cols - 1;
			if (h >= img_color.rows)
				h = img_color.rows - 1;
			glm::vec3 average = glm::vec3(0.0f, 0.0f, 0.0f);
			int nb_point = 0;
			for (int i = x; i <= w; ++i)
			{
				for (int j = y; j <= h; ++j)
				{
					cv::Vec3f color = img_color[j][i];
					average += glm::vec3(color[0], color[1], color[2]);
					++nb_point;
				}
			}
			return average * (1.0f / nb_point);
		}

	//this function is required, it is used by the system to initialize the DPP
	//the variables img_grad, img_color and outut_base_name should be initilized
	static void ini_from_file(const std::string base_name, const std::string& full_name)
	{
		cv::Mat ini_input = cv::imread(full_name);
		cv::Mat input;
		cv::Mat input_grad_x;
		cv::Mat input_grad_y;
		cv::Mat_<float> grad_x;
		cv::Mat_<float> grad_y;

		outut_base_name = base_name;

		ini_input.convertTo(img_color, CV_32FC3);
		img_color *= (1.0 / 255);
		cv::cvtColor(ini_input, input, CV_RGB2GRAY);
		img_proba_label_1 = cv::Mat_<float>(img_color.rows, img_color.cols, 0.0f);
		cv::Sobel(input, input_grad_x, CV_16S, 1, 0, 3);
		cv::Sobel(input, input_grad_y, CV_16S, 0, 1, 3);
		img_grad = cv::Mat_<float>(input_grad_y.rows, input_grad_y.cols);
		input_grad_x.convertTo(grad_x, CV_32FC1);
		input_grad_y.convertTo(grad_y, CV_32FC1);
		input_grad_x *= (1.0 / 255);
		input_grad_y *= (1.0 / 255);
		float* ptr_grad_x = grad_x[0];
		float* ptr_grad_y = grad_y[0];
		float* ptr_grad_norm = img_grad[0];
		float max = std::numeric_limits<float>::min();
		for (int col = 0; col < grad_x.cols; ++col)
		{
			for (int row = 0; row < grad_x.rows; ++row)
			{

				const float norm = *ptr_grad_norm = sqrt(*ptr_grad_x**ptr_grad_x + *ptr_grad_y**ptr_grad_y);
				if (norm > max)
					max = norm;
				++ptr_grad_x;
				++ptr_grad_y;
				++ptr_grad_norm;
			}
		}
		if (max != 0)
		{
			img_grad *= 1.0 / max;// sqrt(1.0 / max);// sqrt(max);
		}
		cv::GaussianBlur(img_grad, img_grad, { 7, 7 }, 4, 4);
	}

	//this function is required, it is used by the system to save some results
	template<typename SA>
	static void save_output(const SA& sa, const int rows, const int cols)
	{
		cv::imwrite((outut_base_name + "_input.png").c_str(), img_color * 255.0f);
		save_graph_label_facet(outut_base_name, "_output.svg", sa.triangulation, rows, cols);
		save_graph_label_facet(outut_base_name, "_output_overlap.svg", sa.triangulation, rows, cols, true, true);
		save_graph_label_facet(outut_base_name, "_output_line.svg", sa.triangulation, rows, cols, false);
		sa.save_energy((outut_base_name + "_energy.csv").c_str());
	}


	//this function is required, it is called by the system at the begining. You can load specific parameters for your application
	static void load_app_specific_parameters()
	{
		std::string str, str2;
		std::istringstream iss;
		std::ifstream ifs("config/app_object_contouring_parameters.conf");
		while (std::getline(ifs, str2).good())
		{
			iss = std::istringstream();//otherwise iss  is not updated (or in a bad state after finisshing to read a line)
			iss.str(str2.c_str());
			std::getline(iss, str, ' ');
			if (str == "alpha")
				iss >> alpha;
			else if (str == "min_length_edge")
				iss >> min_length_edge;
			else if (str == "mu")
				iss >> mu;
		}
		ifs.close();
	}
};

#endif  __HEADER___COST_APPLI_OBJECT_CONTOURING__

