#ifndef __RECUIT__HEADER____
#define __RECUIT__HEADER____

#include"gl_widget.h"

#include <vector>
#include <random>
namespace DPP
{
	enum Sampling_mode { UNIFORM, GUIDED};
	enum Active_kernel { BD_M_L, BD_M, BD_L, BD, B_L, D_L, M_L, B, D, M, L };
	struct Parameters
	{
		DPP::Sampling_mode sampling_mode;
		DPP::Active_kernel active_kernel;
		//if sampling_mode is uniform, we use sampling domain to have the size of the domain, otherwise we use point_distribution
		//to override the values of sampling domain. In all case, we assume that sampling_domain.x = sampling_domain.y = 0 
		glm::vec4 sampling_domain;// = { x, y, cols, rows }
		cv::Mat_<float>* point_distribution;
		int expected_number_of_point;//ie lambda in equation 5 of our paper.
									 //T0 and alpha are the two parameters of equation 6 of our paper. We have T_t = T_0.\alpha^t
		double T_0; // initial temperatiure for simulated annealing. See equation 6 in our paper
		double alpha;//the parameter for the geometric decrease of the temperature. See equation 6 in our paper
		int number_of_maximum_iteration;//the algorithm will run until the number_of_maximum_iteration is reached.
		Parameters()
		{
			load_config();
		}
		//Parameters(DPP::Sampling_mode sm = DPP::Sampling_mode::GUIDED,
		//	DPP::Active_kernel ak = DPP::Active_kernel::BD_M_L,
		//	glm::vec4 sd = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f))
		//	:sampling_mode(sm), active_kernel(ak), sampling_domain(sd), point_distribution(nullptr),
		//	expected_number_of_point(1000),
		//	T_0(0.01),alpha(0.99998000), number_of_maximum_iteration(2000000){}
		bool is_in_sampling_domain(int x, int y)
		{
			return x >= sampling_domain.x && y >= sampling_domain.y && x < sampling_domain.z && y < sampling_domain.w;
		}
		void load_config()
		{
			std::string str, str2;
			std::istringstream iss;
			std::ifstream ifs("config/DPP_parameters.conf");
			while (std::getline(ifs, str2).good())
			{
				iss = std::istringstream();//otherwise iss  is not updated (or in a bad state after finisshing to read a line)
				iss.str(str2.c_str());
				std::getline(iss, str, ' ');
				if (str == "sampling_mode")
				{
					iss >> str;
					if (str == "UNIFORM")
						sampling_mode = DPP::Sampling_mode::UNIFORM;
					else if (str == "GUIDED")
						sampling_mode = DPP::Sampling_mode::GUIDED;
					else
						sampling_mode = DPP::Sampling_mode::UNIFORM;
				}
				else if (str == "active_kernel")
				{
					iss >> str;
					if (str == "BD_M_L")
						active_kernel = DPP::Active_kernel::BD_M_L;
					else if (str == "BD_M")
						active_kernel = DPP::Active_kernel::BD_M;
					else if (str == "BD_L")
						active_kernel = DPP::Active_kernel::BD_L;
					else if (str == "BD")
						active_kernel = DPP::Active_kernel::BD;
					else if (str == "B_L")
						active_kernel = DPP::Active_kernel::B_L;
					else if (str == "D_L")
						active_kernel = DPP::Active_kernel::D_L;
					else if (str == "M_L")
						active_kernel = DPP::Active_kernel::M_L;
					else if (str == "B")
						active_kernel = DPP::Active_kernel::B;
					else if (str == "D")
						active_kernel = DPP::Active_kernel::D;
					else if (str == "M")
						active_kernel = DPP::Active_kernel::M;
					else if (str == "L")
						active_kernel = DPP::Active_kernel::L;
					else
						active_kernel = DPP::Active_kernel::BD_M_L;
				}
				else if (str == "sampling_domain")
					iss >> sampling_domain.x >> sampling_domain.y >> sampling_domain.z >> sampling_domain.w;
				else if (str == "expected_number_of_point")
					iss >> expected_number_of_point;
				else if (str == "T_0")
					iss >> T_0;
				else if (str == "alpha")
					iss >> alpha;
				else if (str == "number_of_maximum_iteration")
					iss >> number_of_maximum_iteration;

			}
			ifs.close();
		}
	};
};
template<typename Point_2>
struct image_distribution
{
	int cols;
	std::discrete_distribution<int> distrib;
	std::uniform_real_distribution<double> perturbation;
	void reset_distribution(DPP::Parameters& param)
	{
		distrib.reset();
		std::vector<double> weight;
		if (param.sampling_mode == DPP::Sampling_mode::GUIDED)
		{
			const cv::Mat_<float>& img = *param.point_distribution;
			param.sampling_domain.x = 0.0f;
			param.sampling_domain.y = 0.0f;
			param.sampling_domain.z = img.cols;
			param.sampling_domain.w = img.rows;
			cols = img.cols;
			weight.reserve(img.rows*img.cols);
			for (int row = 0; row < img.rows; ++row)
			{
				for (int col = 0; col < cols; ++col)
				{
					//we add epsilon (=0.01) to have a non null probability on uniform area
					weight.emplace_back(fabs(img[row][col]) + 0.01);
				}
			}
		}
		else
		{
			weight.resize(param.sampling_domain.z*param.sampling_domain.w, 0.1);
			cols = param.sampling_domain.z;
		}
		//probabilities automatically renormalized by the constructor of discrete_distribution
		distrib = std::discrete_distribution<int>(weight.begin(), weight.end());
		perturbation = std::uniform_real_distribution<double>(0, 0.01);
	}
	template<typename Generator>
	Point_2 operator()(Generator& g)
	{
		const int tmp = distrib(g);
		//the initial idea behind the perturbation is to reduce the probabilty to sample squares or more generally distribution of points which do not correspond to a uniq delaunay triangulation
		//we also move the random point to the center of the pixel (+0.5, +0.5)
		return Point_2(tmp%cols + 0.5 + perturbation(g), tmp / cols + 0.5 + perturbation(g));
	}
	template<typename Generator>
	Point_2 no_perturbation(Generator& g)
	{
		const int tmp = distrib(g);
		return Point_2(tmp%cols + 0.5, tmp / cols + 0.5);
	}
	template<typename Generator>
	Point_2 add_perturbation(Point_2& p1, Generator& g)
	{
		return Point_2(p1.x() + perturbation(g), p1.y() + perturbation(g));
	}
};

template<typename Triangulation_>
struct simulated_annealing
{
	typedef Triangulation_ Triangulation;
	DPP::Parameters param;
	std::random_device rd;
	std::mt19937 gen;
	Triangulation triangulation;
	std::vector<float> energy_iter;
	double T;
	int nb_it;

	int img_cols;
	int img_rows;

	double current_energy;
	double threshold_delta_energy;
	double specific_th_bd;


	std::uniform_int_distribution<int> op_selector;
	std::uniform_real_distribution<double> threshold_selector;
	image_distribution<typename Triangulation::Point> coord_distrib;
	simulated_annealing(DPP::Parameters _param) : param(_param),
		rd(), gen(rd()),
		nb_it(0),
		threshold_selector(0.0, 1.0), op_selector(0, 59),//2*3*4*5*6 eq to 4*3*5 ... because divisible by 2, 3, 4, 5, 6
		T(param.T_0)
	{
		Triangulation::Ini_data_cost::load_app_specific_parameters();
	}
	inline void update_target_nb_point(const int new_number_of_point)
	{
		Triangulation::Ini_data_cost::target_nb_point = new_number_of_point;
		param.expected_number_of_point = new_number_of_point;
		//target_nb_point = new_number_of_point;
	}

	inline void save_energy(const char* file_name)const
	{
		std::ofstream ofs;
		ofs.open(file_name);
		for (int i = 0; i < energy_iter.size(); ++i)
			ofs << i << ";" << energy_iter[i] << std::endl;
		ofs.close();
	}
	void add_random_point()
	{
		for (int i = 0; i < 1000; ++i)
		{
			typename Triangulation::Point point = coord_distrib(gen);
		//	typename Triangulation::Point point = coord_distrib.no_perturbation(gen);
			Triangulation::Locate_type lt;
			int li;
			Triangulation::Face_handle fh = triangulation.locate(point, lt, li);
			if (lt == Triangulation::VERTEX)
				continue;
			if (lt == Triangulation::FACE)
			{
				const typename Triangulation::Vertex_handle v1 = fh->vertex(0);
				const typename Triangulation::Vertex_handle v2 = fh->vertex(1);
				const typename Triangulation::Vertex_handle v3 = fh->vertex(2);
				const auto& p1 = v1->point();
				const auto& p2 = v2->point();
				const auto& p3 = v3->point();
				if (sqrt((p1 - point).squared_length()) < 0.5)
					continue;
				if (sqrt((p2 - point).squared_length()) < 0.5)
					continue;
				if (sqrt((p3 - point).squared_length()) < 0.5)
					continue;
			}
			Triangulation::Vertex_handle new_vert = triangulation.insert(point);
			typename Triangulation::Face_circulator c = new_vert->incident_faces();
			bool is_good = true;
			typename Triangulation::Face_circulator c_end = c;
			do
			{
				const typename Triangulation::Point pt1 = triangulation.dual(c);
				for(int kkk = 0; kkk < 3; ++kkk)
				{
					const typename Triangulation::Point pt2 = triangulation.dual(c->neighbor(kkk));
					if (pt2 == pt1)
					{
						std::cout << "bad triangulation " << std::endl;
						is_good = false;
					}
				}
			} while (++c != c_end);
			if (is_good)
				new_vert->info() = next_info.vertex();
			else
				triangulation.remove(new_vert);

		}
		triangulation.ini_data_cost();
		current_energy = triangulation.get_energy();
	}
	void reset(DPP::Parameters& _param)
	{
		param = _param;
		coord_distrib.reset_distribution(param);
		img_cols = param.sampling_domain.z;
		img_rows = param.sampling_domain.w;
		update_target_nb_point(param.expected_number_of_point);

		nb_it = 0;
		T = param.T_0;

		triangulation.clear();
		next_info.reset();
		triangulation.ini(img_cols, img_rows);

		current_energy = triangulation.get_energy();
		add_random_point();
		std::cout << "compute energy" << std::endl;
		current_energy = triangulation.get_energy();
		std::cout << current_energy << std::endl;
		energy_iter.clear();
		energy_iter.reserve(param.number_of_maximum_iteration);
	}
	
	
	void operator_death()
	{
		if (triangulation.number_of_vertices() == 0)
			return;
		auto it = std::next(triangulation.vertices_begin(), std::uniform_int_distribution<decltype(triangulation.number_of_vertices())>(0, triangulation.number_of_vertices() - 1)(gen));
		const auto pt = it->point();
		if (pt.x() < 0 || pt.x() >= img_cols)
			return;
		specific_th_bd = T * log((triangulation.number_of_vertices()*1.0/ Triangulation::Ini_data_cost::target_nb_point));
		current_energy += (double)triangulation.remove_if_energy_good(it, threshold_delta_energy, specific_th_bd);
		return;
	}
	void operator_birth()
	{
		const typename Triangulation::Point point = coord_distrib(gen);
		if (point.x() < 0 || point.y() < 0 || point.x() >= img_cols || point.y() >= img_rows)
		{
			operator_birth();
			return;
		}
		Triangulation::Locate_type lt;
		int li;
		Triangulation::Face_handle fh = triangulation.locate(point, lt, li);
		if (lt == Triangulation::VERTEX)
			return;
		if (lt == Triangulation::FACE)
		{
			const typename Triangulation::Vertex_handle v1 = fh->vertex(0);
			const typename Triangulation::Vertex_handle v2 = fh->vertex(1);
			const typename Triangulation::Vertex_handle v3 = fh->vertex(2);
			const auto& p1 = v1->point();
			const auto& p2 = v2->point();
			const auto& p3 = v3->point();
			if (sqrt((p1 - point).squared_length()) < 0.5)
				return;
			if (sqrt((p2 - point).squared_length()) < 0.5)
				return;
			if (sqrt((p3 - point).squared_length()) < 0.5)
				return;
		}
		specific_th_bd = T * log(Triangulation::Ini_data_cost::target_nb_point*1.0 / (triangulation.number_of_vertices() + 1));
		current_energy += (double)triangulation.insertion_if_energy_good(point, 0, threshold_delta_energy, specific_th_bd);
		return;
	}
	void operator_label_edge()
	{
		auto it = std::next(triangulation.edges_begin(), std::uniform_int_distribution<decltype(triangulation.number_of_vertices())>(0, triangulation.number_of_vertices() + triangulation.number_of_faces() - 2)(gen));
		current_energy += (double)triangulation.change_label_edge_if_energy_good(*it, threshold_delta_energy, std::uniform_int_distribution<int>(0, 1)(gen));// 4 - 1)(gen));

		return;
	}

	void operator_label_partition_vertex()
	{
		auto it = std::next(triangulation.finite_vertices_begin(), std::uniform_int_distribution<decltype(triangulation.number_of_vertices())>(0, triangulation.number_of_vertices() - 1)(gen));
		int nb_vertex_around = 0;
		auto it2 = it->incident_vertices();
		auto it2_end = it2;
		do
		{
			++nb_vertex_around;
		} while (++it2 != it2_end);
		current_energy += (double)triangulation.change_label_vertex_partition_if_energy_good(it, threshold_delta_energy, std::uniform_int_distribution<int>(0, nb_vertex_around)(gen));
		return;
	}
	void operator_label_partition_face()
	{
		auto it = std::next(triangulation.finite_faces_begin(), std::uniform_int_distribution<decltype(triangulation.number_of_faces())>(0, triangulation.number_of_faces() - 1)(gen));
		current_energy += (double)triangulation.change_label_face_partition_if_energy_good(it, threshold_delta_energy, std::uniform_int_distribution<int>(0, 1)(gen));
		return;
	}
	void operator_merge_curve()
	{
		current_energy += triangulation.merge_curve_if_energy_good(threshold_delta_energy);
	}
	void operator_split_curve()
	{
		current_energy += triangulation.split_curve_if_energy_good(threshold_delta_energy);
	}
	void operator_move()
	{
		if (triangulation.number_of_vertices() == 0)
			return;
		auto it = std::next(triangulation.vertices_begin(), std::uniform_int_distribution<decltype(triangulation.number_of_vertices())>(0, triangulation.number_of_vertices() - 1)(gen));
		const typename Triangulation::Point point = it->point();
		if (point.x() < 0 || point.x() >= img_cols)
			return;
		current_energy += (double)triangulation.move_if_energy_good(it, threshold_delta_energy, gen, img_cols, img_rows);
	}
	inline void one_step()
	{
		double threshold = threshold_selector(gen);
		threshold_delta_energy = (threshold == 0.0 ? std::numeric_limits<double>::max() : -T*log(threshold));
#ifdef STANDARD_RANDOM_OPERATOR_SELECTOR
		int rnd = op_selector(gen);
		switch (param.active_kernel)
		{
		case DPP::Active_kernel::BD_M_L:
			switch (rnd % 4)
			{
			case 0:operator_birth(); break;
			case 1:operator_death(); break;
			case 2:operator_move(); break;
			case 3:Triangulation::Ini_data_cost::operator_label(this); break;
			}
			break;
		case DPP::Active_kernel::BD_M:
			switch (rnd % 3)
			{
			case 0:operator_birth(); break;
			case 1:operator_death(); break;
			case 2:operator_move(); break;
			}
			break;
		case DPP::Active_kernel::BD_L:
			switch (rnd % 3)
			{
			case 0:operator_birth(); break;
			case 1:operator_death(); break;
			case 2:Triangulation::Ini_data_cost::operator_label(this); break;
			}
			break;
		case DPP::Active_kernel::BD:
			switch (rnd % 2)
			{
			case 0:operator_birth(); break;
			case 1:operator_death(); break;
				break;
			}
		case DPP::Active_kernel::B_L:
			switch (rnd % 2)
			{
			case 0:operator_birth(); break;
			case 1:Triangulation::Ini_data_cost::operator_label(this); break;
				break;
			}
		case DPP::Active_kernel::D_L:
			switch (rnd % 2)
			{
			case 0:operator_death(); break;
			case 1:Triangulation::Ini_data_cost::operator_label(this); break;
				break;
			}
		case DPP::Active_kernel::M_L:
			switch (rnd % 2)
			{
			case 0:operator_move(); break;
			case 1:Triangulation::Ini_data_cost::operator_label(this); break;
				break;
			}
		case DPP::Active_kernel::B:operator_birth();break;
		case DPP::Active_kernel::D:operator_death();break;
		case DPP::Active_kernel::M:operator_move();break;
		case DPP::Active_kernel::L:Triangulation::Ini_data_cost::operator_label(this);break;
		}
#else
		Triangulation::Ini_data_cost::operator_label(this);
		Triangulation::Ini_data_cost::operator_label(this);
		Triangulation::Ini_data_cost::operator_label(this);
		Triangulation::Ini_data_cost::operator_label(this);
		Triangulation::Ini_data_cost::operator_label(this);
		Triangulation::Ini_data_cost::operator_label(this);
		Triangulation::Ini_data_cost::operator_label(this);
		Triangulation::Ini_data_cost::operator_label(this);
		Triangulation::Ini_data_cost::operator_label(this);
		Triangulation::Ini_data_cost::operator_label(this);
		operator_move();
		operator_move();
		operator_move();
		operator_move();
		operator_move();
		operator_move();
		operator_move();
		operator_move();
		//		if (threshold_selector(gen) < atan((int(triangulation.number_of_vertices()) - Triangulation::Ini_data_cost::target_nb_point)* tan(M_PI*(0.5-0.1)) / target_nb_point)*M_1_PI + 0.5)
		if (T > 0.0000000001)
		{
			operator_birth();
			operator_death();
		}
#endif
		T *= param.alpha;
	}
	inline void all_step()
	{
		current_energy = triangulation.get_energy();
		for (; nb_it < param.number_of_maximum_iteration; ++nb_it)
		{
			one_step();
			if (nb_it % 100 == 0)
				energy_iter.emplace_back(triangulation.get_energy());
			if (nb_it % 10000 == 0)
			{
				std::cout << nb_it<<"\t"<< T << "\t" << param.alpha << "\t" << current_energy << " " << triangulation.number_of_vertices() << std::endl;;
			}
		}
	}
};

#endif

