#pragma once
#ifndef __MAIN_WINDOWS__H__
#define __MAIN_WINDOWS__H__
#include"util_define.h"

#include <QMainWindow>
#include <QMenuBar>
#include<QWidgetAction>
#include <QSpinBox>
#include <QDoubleSpinBox>
#include<QLayout>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include<QGridLayout>
#include<QtOpenGL/QGLWidget>
#include <QSlider>
#include<QGroupBox>
#include<QLabel>
#include<QPushButton>
#include<QComboBox>
#include<QCheckBox>
#include <QColor>
#include <QFileDialog>
#include<QtWidgets/QApplication>
#include<QtWidgets/QMessageBox>

#include"gl_widget.h"
#include"algo_thread.h"
#include"simulated_annealing.h"
template<typename SA>
class MainWindows : public QMainWindow
{
public:
	gl_widget* gl_wgt;
	SA sa;
	thread_algo<SA> algo;
	bool add_random_point;
public:
	MainWindows(QWidget *parent = nullptr) : QMainWindow(parent), sa(DPP::Parameters())
	{
		sa.param.point_distribution = &SA::Triangulation::Ini_data_cost::img_grad;
		SA::Triangulation::Ini_data_cost::target_nb_point = sa.param.expected_number_of_point;
		add_random_point = false;
		QGridLayout* layout;
		QGridLayout *layout_grid = new QGridLayout;
		QGridLayout *layout_bottom_grid = new QGridLayout;
		QGridLayout *layout_hbox = new QGridLayout;
		QWidget* wgt = new QWidget(this);
		QWidget* wgt_tmp;
		gl_wgt = new gl_widget(this);
		q_slider_change_alpha::mon_gl_widget = gl_wgt;
		//setFixedSize(1600, 900);
		setGeometry(100, 126, 1024, 1024);

		layout = new QGridLayout;
		QDoubleSpinBox* double_spin_box;
		QSpinBox* point_number;
		int cpt_layout = 0;
		for (std::pair<std::string, double*>& param_spin_box : SA::Triangulation::Gui::link_double_spinbox_in_gui)
		{
			layout->addWidget(new QLabel(param_spin_box.first.c_str(), this), cpt_layout, 0);

			double_spin_box = new QDoubleSpinBox(this);
			double_spin_box->setMinimum(0);
			double_spin_box->setMaximum(1);
			double_spin_box->setDecimals(7);
			double_spin_box->setSingleStep(0.01);
			double_spin_box->setValue(*param_spin_box.second);
			QObject::connect(double_spin_box, static_cast<void(QDoubleSpinBox::*)(double)>(&QDoubleSpinBox::valueChanged), [=](double new_val) {*param_spin_box.second = new_val; });
			layout->addWidget(double_spin_box, cpt_layout, 1);
			++cpt_layout;
		}
		for (std::pair<std::string, int*>& param_spin_box : SA::Triangulation::Gui::link_int_spinbox_in_gui)
		{
			layout->addWidget(new QLabel(param_spin_box.first.c_str(), this), cpt_layout, 0);

			point_number = new QSpinBox(this);
			point_number->setMinimum(0);
			point_number->setMaximum(100000);
			point_number->setSingleStep(1);
			point_number->setValue(*param_spin_box.second);
			QObject::connect(point_number, static_cast<void(QSpinBox::*)(int)>(&QSpinBox::valueChanged), [=](int new_val) {*param_spin_box.second = new_val; });
			layout->addWidget(point_number, cpt_layout, 1);
			++cpt_layout;
		}

		wgt_tmp = new QWidget(this);
		wgt_tmp->setLayout(layout);
		layout_bottom_grid->addWidget(wgt_tmp, 0, 0);

		QSlider* iteration(new QSlider(Qt::Horizontal, parent));
		iteration->setValue(sa.nb_it);
		iteration->setMaximum(sa.param.number_of_maximum_iteration);
		QObject::connect(iteration, &QSlider::valueChanged, [=](int i) {
			
			sa.nb_it = i;
			sa.T = sa.param.T_0 * pow(sa.param.alpha, i);
			if (algo.end_reach)
			{
				algo.stop();
				algo.launch(sa, gl_wgt);
			}
		});

		layout = new QGridLayout;
		layout->addWidget(new QLabel("High perturbations", this), 0, 0);
		layout->addWidget(iteration, 0, 1,1,3);
		layout->addWidget(new QLabel("Low perturbations", this), 0, 5);
		
		layout->addWidget(new QLabel("Number of expected points", this), 1, 0);
		point_number = new QSpinBox(this);
		point_number->setMinimum(0);
		point_number->setMaximum(100000);
		point_number->setSingleStep(1);
		point_number->setValue(SA::Triangulation::Ini_data_cost::target_nb_point);
		QObject::connect(point_number, static_cast<void(QSpinBox::*)(int)>(&QSpinBox::valueChanged), [=](int new_val) {sa.update_target_nb_point(new_val);});
		layout->addWidget(point_number, 1, 1);

		QPushButton* button = new QPushButton("Add random points", this);
		layout->addWidget(button, 1, 5);
		QObject::connect(button, &QPushButton::pressed, [=]() { add_random_point = true; });
		
		layout->addWidget(new QLabel("nb_it_max", this), 2, 0);
		point_number = new QSpinBox(this);
		point_number->setMinimum(0);
		point_number->setMaximum(std::numeric_limits<int>::max());
		point_number->setValue(sa.param.number_of_maximum_iteration);
		QObject::connect(point_number, static_cast<void(QSpinBox::*)(int)>(&QSpinBox::valueChanged), [=](int new_val) {sa.param.number_of_maximum_iteration = new_val;	iteration->setMaximum(sa.param.number_of_maximum_iteration); });
		layout->addWidget(point_number, 2, 1);

		layout->addWidget(new QLabel("T_0", this), 2, 2);
		double_spin_box = new QDoubleSpinBox(this);
		double_spin_box->setMinimum(0);
		double_spin_box->setMaximum(1000);
		double_spin_box->setSingleStep(0.01);
		double_spin_box->setValue(sa.param.T_0);
		QObject::connect(double_spin_box, static_cast<void(QDoubleSpinBox::*)(double)>(&QDoubleSpinBox::valueChanged), [=](double new_val) {sa.param.T_0 = new_val; });
		layout->addWidget(double_spin_box, 2, 3);
		

		layout->addWidget(new QLabel("dt", this), 2, 4);
		double_spin_box = new QDoubleSpinBox(this);
		double_spin_box->setMinimum(0);
		double_spin_box->setMaximum(1);
		double_spin_box->setDecimals(10);
		double_spin_box->setSingleStep(0.00000001);
		double_spin_box->setValue(sa.param.alpha);
		QObject::connect(double_spin_box, static_cast<void(QDoubleSpinBox::*)(double)>(&QDoubleSpinBox::valueChanged), [=](double new_val){sa.param.alpha = new_val; });
		layout->addWidget(double_spin_box, 2, 5);

		wgt_tmp = new QWidget(this);
		wgt_tmp->setLayout(layout);
		layout_bottom_grid->addWidget(wgt_tmp, 0, 1);
		this->setStyleSheet("QLabel, QWidget { font: 18px \"Times New Roman\";}");
		QWidget* gridGroupBox_2 = new QWidget(this);
		gridGroupBox_2->setLayout(layout_bottom_grid);
		layout_hbox->addWidget(gl_wgt, 0, 0, 9, 1);
		layout_hbox->addWidget(gridGroupBox_2, 10, 0);
		wgt->setLayout(layout_hbox);
		this->setCentralWidget(wgt);

		QMenu* menu;
		QAction* elem;
		QWidgetAction* elem_widget;

		QObject::connect(gl_wgt, &gl_widget::job_end, gl_wgt, [=]() {
			if (add_random_point)
			{
				sa.add_random_point();
				add_random_point = false;
			}
			std::vector<float> alpha;
			alpha.reserve(gl_wgt->v_change_alpha_event.size());
			for (int i = 0; i < (int)gl_wgt->v_change_alpha_event.size(); ++i)
				alpha.emplace_back(*gl_wgt->v_change_alpha_event[i]->__alpha);
		//	float al = *gl_wgt->v_change_alpha_event[4]->__alpha;
			std::vector<float*> p_tmp = SA::Triangulation::Gui::ini_vbo_get_alpha(sa.triangulation, gl_wgt, gl_wgt->vbo, "node", 0.5f, glm::vec3(0.8f, 0.5f, 0.5f), { glm::vec3(0.0f, 0.0f, 1.0f) ,glm::vec3(1.0f, 0.0f, 0.0f) ,glm::vec3(0.0f, 1.0f, 0.0f) });
			for (int i = 0; i<(int)p_tmp.size(); ++i)
				*p_tmp[i] = alpha[2 + i];
			gl_wgt->update();
			iteration->setValue(sa.nb_it);
			std::cout << sa.nb_it << "\t" << sa.param.alpha << "\t" << sa.current_energy
				<< "\t" << sa.triangulation.get_energy() << "\t"
				<< sa.triangulation.number_of_vertices() << "\t"
				<< sa.triangulation.edge_data.size() << "\t"
				<< sa.triangulation.list_of_curve.size() << "\t"
				<< sa.T << "\t" << std::endl;
		}, Qt::BlockingQueuedConnection);
		QObject::connect(gl_wgt, &gl_widget::job_end_finish, gl_wgt, [=]() {
			SA::Triangulation::Ini_data_cost::save_output(sa, SA::Triangulation::Ini_data_cost::img_color.rows, SA::Triangulation::Ini_data_cost::img_color.cols);
		});
		menu = menuBar()->addMenu("File");
		{
			elem = menu->addAction(tr("Open File"));
			elem->setShortcut(Qt::CTRL + Qt::Key_O);
			elem->setStatusTip("open a file");
			QObject::connect(elem, &QAction::triggered, [=](bool){
				algo.stop();
				std::ifstream ifs("config/default_dir_open.conf");
				std::string tmp_path_open;
				ifs >> tmp_path_open;
				const QString& qt_full_name_file = QFileDialog::getOpenFileName(this->parentWidget(), tr("Open File"),
					tmp_path_open.c_str(), tr("Images (*.png *.jpg *.tif *.tiff *.JPEG *.bmp);;All (*.*)"));
				if (qt_full_name_file == "")
					return;

				SA::Triangulation::Ini_data_cost::ini_from_file(QFileInfo(qt_full_name_file).completeBaseName().toStdString(), qt_full_name_file.toStdString());
			cv::imwrite("input_grad_norm.png", SA::Triangulation::Ini_data_cost::img_grad*255.0f);
				gl_wgt->set_input_img(SA::Triangulation::Ini_data_cost::img_color);
				gl_wgt->v_change_alpha_event[1]->_connect(&gl_wgt->ini_img(SA::Triangulation::Ini_data_cost::img_grad, "input_grad_norm", 0.7f)->alpha);
				gl_wgt->update();
				gl_wgt->pixel_label_1.clear();
				gl_wgt->pixel_label_2.clear();
				sa.reset(sa.param);
				std::vector<float*> p_tmp = SA::Triangulation::Gui::ini_vbo_get_alpha(sa.triangulation, gl_wgt,gl_wgt->vbo, "node", 0.5f, glm::vec3(0.8f, 0.5f, 0.5f), { glm::vec3(0.0f, 0.0f, 1.0f) ,glm::vec3(1.0f, 0.0f, 0.0f),glm::vec3(0.0f, 1.0f, 0.0f) });
				for(int i=0; i<(int)p_tmp.size(); ++i)
					gl_wgt->v_change_alpha_event[2 + i]->_connect(p_tmp[i]);			
				algo.launch(sa, gl_wgt);

			});

			elem = menu->addAction(tr("Recompute proba"));
			elem->setShortcut(Qt::CTRL + Qt::Key_T);
			elem->setStatusTip("open a file");
			QObject::connect(elem, &QAction::triggered, [=](bool) {
				SA::Triangulation::Ini_data_cost::process_user_classes_input(gl_wgt->pixel_label_1, gl_wgt->pixel_label_2);
			});
			elem = menu->addAction(tr("Save output"));
			elem->setShortcut(Qt::CTRL + Qt::Key_S);
			elem->setStatusTip(tr("save output"));
			QObject::connect(elem, &QAction::triggered, [=](bool){	
				SA::Triangulation::Ini_data_cost::save_output(sa, SA::Triangulation::Ini_data_cost::img_color.rows, SA::Triangulation::Ini_data_cost::img_color.cols);
			});

			elem = menu->addAction(tr("Exit"));
			elem->setShortcut(Qt::Key_Escape);
			elem->setStatusTip(tr("quit the application"));
			QObject::connect(elem, (&QAction::triggered), gl_wgt, [=](bool){
				algo.stop();
				qApp->quit();
			}, Qt::QueuedConnection);
		}
		std::list<std::pair<std::string, QGridLayout*>> lst_menu;
		{
			layout = new QGridLayout;
			lst_menu.emplace_back("Show image", layout);
			int i = 1;
			std::vector<std::string> v_str = { "input_img", "img_grad_norm"};
			for (const std::string& str : SA::Triangulation::Gui::extra_alpha_parameter)
				v_str.emplace_back(str);
			for (const std::string& str : v_str)
			{
				q_slider_change_alpha* tmp = new q_slider_change_alpha(this);
				tmp->setValue(255);
				gl_wgt->v_change_alpha_event.push_back(tmp);
				layout->addWidget(new QLabel(str.c_str(), this), i, 0);
				layout->addWidget(tmp, i, 1);
				++i;
			}
		}
		for (std::pair<std::string, QGridLayout*>& p : lst_menu)
		{
			menu = menuBar()->addMenu(p.first.c_str());
			elem_widget = new QWidgetAction(this);
			QWidget* wgt = new QWidget(menu);
			wgt->setLayout(p.second);
			elem_widget->setDefaultWidget(wgt);
			menu->addAction(elem_widget);
		}
	}
	virtual ~MainWindows()
	{
	//	delete mw;
	}
};


#endif
