
#include "Texture.hpp"
//#include "Debug.h"


base_texture::base_texture(const glm::vec2& size)
{
	//mTextureId = -1;
	mTextureType = GL_TEXTURE_2D;
	glGenTextures(1,&mTextureId);
	glBindTexture(GL_TEXTURE_2D, mTextureId); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);//GL_NEAREST
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	//unsigned char* data = new unsigned char[int(size.x) * int(size.y) * 4];
	//for (int i = 0,i_end((int)floor(size.x*size.y*4));i!=i_end; i++)
	//	data[i] = 0;
	//glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (GLsizei)size.x, (GLsizei)size.y, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (GLsizei)size.x, (GLsizei)size.y, 0, GL_RGBA, GL_UNSIGNED_BYTE, nullptr);
	//delete[] data;
}
base_texture::base_texture() {
	mTextureId = -1;
	mTextureType = GL_TEXTURE_2D;
}

base_texture::base_texture(const std::string& path) {
	mTextureId = -1;
	mTextureType = GL_TEXTURE_2D;
	loadTexture(path);
}

base_texture::base_texture(const GLuint id) {
	mTextureId = id;
	mTextureType = GL_TEXTURE_2D;
	std::stringstream ss;
	ss << "Texture from id: " << id;
	mName = ss.str();
}

base_texture::~base_texture() {
	glDeleteTextures(1, &mTextureId);
}
	
/**
 * @brief Binds the texture.
 * It has to be done before glBegin!
 */
void base_texture::bind() {
	glBindTexture(GL_TEXTURE_2D, mTextureId);
}

GLuint base_texture::getTextureId() const {
	return mTextureId;
}
bool base_texture::download_texture(const std::string& name)
{
	GLint w,h;
	glBindTexture(GL_TEXTURE_2D, mTextureId);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &w);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &h);
	//unsigned char* data = new unsigned char[w*h*4];
	//glGetTexImage(GL_TEXTURE_2D,0,GL_RGBA,GL_UNSIGNED_BYTE,data);
	unsigned char* data = new unsigned char[w*h*4];
	glGetTexImage(GL_TEXTURE_2D,0,GL_BGRA,GL_UNSIGNED_BYTE,data);
	bool tmp = save_texture(name,data,w,h);
	delete[] data;
	return tmp;
}
void base_texture::uploadTexture(const void* data, size_t size_x, size_t size_y, GLuint src_type)
{
	if (data != 0) {
		glBindTexture(GL_TEXTURE_2D, mTextureId); //we activate this id to tell opengl we are going to use this texture
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		/**
		 * To eliminate white edge effect
		 **/
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		// After this, there is no need to keep track of the image, it has been uploaded to the GPU
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (GLsizei)size_x,
			(GLsizei)size_y, 0, src_type, GL_UNSIGNED_BYTE,
				data);
	} else {
		std::cerr << "Invalid image " << mName << std::endl;
	}
}
